import logging
from enum import Enum
from io import IOBase
from pathlib import PurePath
from typing import Union

from pdfminer.high_level import extract_pages

from hotpdf.memory_map import MemoryMap

logging.getLogger("pdfminer").setLevel(logging.ERROR)


class Result(Enum):
    LOADED = 0
    LOCKED = 1
    WRONG_PASSWORD = 2
    UNKNOWN_ERROR = 3


def __process(
    source: Union[PurePath, str, IOBase],
    password: str = "",
    drop_duplicate_spans: bool = True,
    first_page: int = 0,
    last_page: int = 0,
) -> list[MemoryMap]:
    pages: list[MemoryMap] = []
    page_numbers = None if first_page == 0 and last_page == 0 else list(range(first_page - 1, last_page))
    hl_page_layouts = extract_pages(source, password=password, page_numbers=page_numbers, caching=False)
    for page_layout in hl_page_layouts:
        parsed_page: MemoryMap = MemoryMap()
        parsed_page.build_memory_map()
        parsed_page.load_memory_map(page=page_layout, drop_duplicate_spans=drop_duplicate_spans)
        pages.append(parsed_page)
    return pages


def process(
    source: Union[PurePath, str, IOBase],
    password: str = "",
    drop_duplicate_spans: bool = True,
    first_page: int = 0,
    last_page: int = 0,
) -> list[MemoryMap]:
    return __process(
        source=source,
        password=password,
        drop_duplicate_spans=drop_duplicate_spans,
        first_page=first_page,
        last_page=last_page,
    )



import time
import tracemalloc

import pytest

from hotpdf import HotPdf


def perform_speed_test(file_name, expected_processing_seconds):
    start_time = time.time()
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(file_name)
    end_time = time.time()
    elapsed = end_time - start_time
    assert (elapsed) < expected_processing_seconds, "Benchmark time exceeded!"


@pytest.mark.skip(reason="Need to perform benchmarks first for pdfminer")
def perform_memory_test(file_name, expected_peak_memory):
    tracemalloc.start()
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(file_name)
    peak_memory = tracemalloc.get_traced_memory()[1] / (1024 * 1024)
    tracemalloc.stop()
    assert peak_memory < expected_peak_memory, "Benchmark memory usage exceeded!"


def test_speed_benchmark_multiple_pages(multiple_pages_file_name):
    perform_speed_test(multiple_pages_file_name, 3.5)


def test_memory_benchmark_multiple_pages(multiple_pages_file_name):
    perform_memory_test(multiple_pages_file_name, 16)


def test_speed_luca_mock(mock_hotpdf_bank_file_name):
    perform_speed_test(mock_hotpdf_bank_file_name, 3.5)


def test_memory_luca_mock(mock_hotpdf_bank_file_name):
    perform_memory_test(mock_hotpdf_bank_file_name, 12.99)


def test_speed_default_file(valid_file_name):
    perform_speed_test(valid_file_name, 3.5)


@pytest.mark.skip(reason="Need to perform benchmarks first")
def test_memory_default_file(valid_file_name):
    perform_memory_test(valid_file_name, 1)



"""Implementation from: https://github.com/puyuan/py-nanoid."""

from math import ceil, log
from os import urandom
from typing import Callable


def algorithm_generate(random_bytes: int) -> bytearray:
    """Algorithm to generate random bytes.

    Args:
        random_bytes (int):

    Returns:
        bytearray:
    """
    return bytearray(urandom(random_bytes))


def method(algorithm: Callable[[int], bytearray], alphabet: str, size: int) -> str:
    alphabet_len = len(alphabet)

    mask = 1
    if alphabet_len > 1:
        mask = (2 << int(log(alphabet_len - 1) / log(2))) - 1
    step = ceil(1.6 * mask * size / alphabet_len)

    id = ""
    while True:
        random_bytes = algorithm(step)

        for i in range(step):
            random_byte = random_bytes[i] & mask
            if random_byte < alphabet_len and alphabet[random_byte]:
                id += alphabet[random_byte]

                if len(id) == size:
                    return id


def generate_nano_id(
    alphabet: str = "_-0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
    size: int = 21,
) -> str:
    return method(algorithm_generate, alphabet, size)



from collections.abc import Iterable
from typing import Union

from .data.classes import HotCharacter, Span


class SpanMap:
    """Hashmap to store spans and their child words for fast referencing
    and character grouping.

    Keys are span_ids and values are Span objects.
    """

    def __init__(self) -> None:
        self.span_map: dict[str, Span] = dict()

    def __len__(self) -> int:
        return len(self.span_map)

    def __getitem__(self, span_id: str) -> Union[Span, None]:
        return self.get_span(span_id)

    def __setitem__(self, span_id: str, hot_character: HotCharacter) -> None:
        self.insert(span_id, hot_character)

    def items(self) -> Iterable[tuple[str, Span]]:
        yield from self.span_map.items()

    def insert(self, span_id: str, hot_character: HotCharacter) -> None:
        span = self.span_map.get(span_id)
        if not span:
            span = Span(
                characters=[],
                span_id=span_id,
            )
        span.characters.append(hot_character)
        self.span_map[span_id] = span

    def get_span(self, span_id: str) -> Union[Span, None]:
        return self.span_map.get(span_id)



import math
import os
from collections import defaultdict
from io import IOBase
from pathlib import PurePath
from typing import Optional, Union

from hotpdf import processor
from hotpdf.memory_map import MemoryMap
from hotpdf.utils import filter_adjacent_coords, intersect

from .data.classes import ElementDimension, HotCharacter, PageResult, SearchResult, Span


class HotPdf:
    def __init__(
        self,
        pdf_file: Union[PurePath, str, IOBase, None] = None,
        password: str = "",
        drop_duplicate_spans: bool = True,
        first_page: int = 0,
        last_page: int = 0,
        extraction_tolerance: int = 4,
    ) -> None:
        """Initialize the HotPdf class.

        Args:
            pdf_file (PurePath | str | IOBytes): The path to the PDF file to be loaded, or a bytes object.
            password (str, optional): Password to use to unlock the pdf
            drop_duplicate_spans (bool, optional): Drop duplicate spans when loading. Defaults to True.
            first_page (int, optional): The first page to load (1-indexed). Defaults to 0 (will load full PDF).
            last_page (int, optional): The last page to load (1-indexed). Defaults to 0 (will load full PDF).
            extraction_tolerance (int, optional): Tolerance value used during text extraction
                to adjust the bounding box for capturing text. Defaults to 4.

        Raises:
            ValueError: If the page range is invalid.
            FileNotFoundError: If the file is not found.
            PermissionError: If the file is encrypted or the password is wrong.
            RuntimeError: If an unkown error is generated by transfotmer.
        """
        self.pages: list[MemoryMap] = []
        self.extraction_tolerance: int = extraction_tolerance
        if pdf_file:
            self.load(pdf_file, password, drop_duplicate_spans, first_page, last_page)

    def __check_file_exists(self, pdf_file: str) -> None:
        if not os.path.exists(pdf_file):
            raise FileNotFoundError(f"File {pdf_file} not found")

    def __check_coordinates(self, x0: int, y0: int, x1: int, y1: int) -> None:
        if x0 < 0 or x1 < 0 or y0 < 0 or y1 < 0:
            raise ValueError("Invalid coordinates")

    def __check_page_number(self, page: int) -> None:
        if page < 0 or page >= len(self.pages):
            raise ValueError("Invalid page number")

    def __check_page_range(self, first_page: int, last_page: int) -> None:
        if first_page > last_page or first_page < 0 or last_page < 0:
            raise ValueError("Invalid page range")

    def __prechecks(self, pdf_file: Union[PurePath, str, IOBase], first_page: int, last_page: int) -> None:
        if type(pdf_file) is str:
            self.__check_file_exists(pdf_file)
        self.__check_page_range(first_page, last_page)

    def load(
        self,
        pdf_file: Union[PurePath, str, IOBase],
        password: str = "",
        drop_duplicate_spans: bool = True,
        first_page: int = 0,
        last_page: int = 0,
    ) -> None:
        """Load a PDF file into memory.

        Args:
            pdf_file (str | Bytes): The path to the PDF file to be loaded, or a bytes object.
            password (str, optional): Password to use to unlock the pdf
            drop_duplicate_spans (bool, optional): Drop duplicate spans when loading. Defaults to True.
            first_page (int, optional): The first page to load. Defaults to 0.
            last_page (int, optional): The last page to load. Defaults to 0.

        Raises:
            Exception: If an unknown error is generated by pdfminer.
        """
        self.__prechecks(pdf_file, first_page, last_page)
        try:
            self.pages = processor.process(pdf_file, password, drop_duplicate_spans, first_page, last_page)
        except Exception as e:
            raise e

    def __extract_full_text_span(
        self,
        hot_characters: list[HotCharacter],
        page_num: int,
    ) -> Union[list[HotCharacter], None]:
        """Extract the full span of text that the given hot characters are a part of.

        Args:
            hot_characters (list[HotCharacter]): the list of hot characters to extract the span for.
            page_num (int): the page number of the hot characters.

        Returns:
            Union[list[HotCharacter], None]: the full span of text that the hot characters are a part of.
        """
        _span: Optional[Span] = None
        if hot_characters[0].span_id:
            _span = self.pages[page_num].span_map[hot_characters[0].span_id]
        return _span.characters if _span else None

    def find_text(
        self,
        query: str,
        pages: Optional[list[int]] = None,
        validate: bool = True,
        take_span: bool = False,
        sort: bool = True,
    ) -> SearchResult:
        """Find text within the loaded PDF pages.

        Args:
            query (str): The text to search for.
            pages (list[int], optional): List of page numbers to search.
            validate (bool, optional): Double check the extracted bounding boxes with the query string.
            take_span (bool, optional): Take the full span of the text that it is a part of.
            sort (bool, Optional): Return elements sorted by their positions.
        Raises:
            ValueError: If the page number is invalid.

        Returns:
            SearchResult: A dictionary mapping page numbers to found text coordinates.
        """
        if pages is None:
            pages = []

        for page in pages:
            self.__check_page_number(page)

        query_pages = (
            {i: self.pages[i] for i in range(len(self.pages))} if len(pages) == 0 else {i: self.pages[i] for i in pages}
        )

        found_page_map = {}

        for page_num in query_pages:
            found_page_map[page_num] = filter_adjacent_coords(*query_pages[page_num].find_text(query))

        final_found_page_map: SearchResult = defaultdict(PageResult)

        for page_num in found_page_map:
            hot_character_page_occurences: PageResult = found_page_map[page_num]
            final_found_page_map[page_num] = []
            seen_span_ids: list[str] = []
            for hot_characters in hot_character_page_occurences:
                text = "".join(hc.value for hc in hot_characters)
                if (query in text) or not validate:
                    full_span_dimension_hot_characters: Union[list[HotCharacter], None] = (
                        self.__extract_full_text_span(
                            hot_characters=hot_characters,
                            page_num=page_num,
                        )
                        if take_span
                        else None
                    )
                    chars_to_append = (
                        full_span_dimension_hot_characters
                        if (take_span and full_span_dimension_hot_characters)
                        else hot_characters
                    )
                    if chars_to_append:
                        if chars_to_append[0].span_id in seen_span_ids:
                            continue
                        seen_span_ids.extend(list(set([ch.span_id for ch in chars_to_append])))
                        if sort:
                            chars_to_append = sorted(chars_to_append, key=lambda ch: (ch.x, ch.y))
                    final_found_page_map[page_num].append(chars_to_append)
        return final_found_page_map

    def extract_spans(self, x0: int, y0: int, x1: int, y1: int, page: int = 0) -> list[Span]:
        """Extract spans that intersect with the given bounding box.

        Args:
            x0 (int): The left x-coordinate of the bounding box.
            y0 (int): The bottom y-coordinate of the bounding box.
            x1 (int): The right x-coordinate of the bounding box.
            y1 (int): The top y-coordinate of the bounding box.
            page (int, optional): The page number. Defaults to 0.
            sort (bool, optional): Sort the spans by their coordinates. Defaults to True.

        Raises:
            ValueError: If the coordinates are invalid.
            ValueError: If the page number is invalid.

        Returns:
            list[Span]: List of spans of hotcharacters that intersect with the given bounding box
        """
        spans: list[Span] = []

        self.__check_coordinates(x0, y0, x1, y1)
        self.__check_page_number(page)

        for _, span in self.pages[page].span_map.items():
            if intersect(ElementDimension(x0, y0, x1, y1, ""), span.get_element_dimension()):
                spans.append(span)
        return spans

    def extract_text(
        self,
        x0: int,
        y0: int,
        x1: int,
        y1: int,
        page: int = 0,
    ) -> str:
        """Extract text from a specified bounding box on a page.

        Args:
            x0 (int): The left x-coordinate of the bounding box.
            y0 (int): The bottom y-coordinate of the bounding box.
            x1 (int): The right x-coordinate of the bounding box.
            y1 (int): The top y-coordinate of the bounding box.
            page (int): The page number. Defaults to 0.

        Raises:
            ValueError: If the coordinates are invalid.
            ValueError: If the page number is invalid.

        Returns:
            str: Extracted text within the bounding box.
        """
        self.__check_coordinates(x0, y0, x1, y1)
        self.__check_page_number(page)

        page_to_search: MemoryMap = self.pages[page]
        extracted_text = page_to_search.extract_text_from_bbox(
            x0=math.floor(x0),
            x1=math.ceil(x1 + self.extraction_tolerance),
            y0=y0,
            y1=y1,
        )
        return extracted_text

    def extract_page_text(
        self,
        page: int,
    ) -> str:
        """Extract text from a specified page.

        Args:
            page (int): The page number.

        Raises:
            ValueError: If the page number is invalid.

        Returns:
            str: Extracted text from the page.
        """
        self.__check_page_number(page)

        page_to_search: MemoryMap = self.pages[page]
        extracted_text = page_to_search.extract_text_from_bbox(
            x0=0,
            x1=page_to_search.width,
            y0=0,
            y1=page_to_search.height,
        )
        return extracted_text

    def extract_spans_text(
        self,
        x0: int,
        y0: int,
        x1: int,
        y1: int,
        page: int = 0,
    ) -> str:
        """Extract text from spans that intersect with the given bounding box.

        Args:
            x0 (int): The left x-coordinate of the bounding box.
            y0 (int): The bottom y-coordinate of the bounding box.
            x1 (int): The right x-coordinate of the bounding box.
            y1 (int): The top y-coordinate of the bounding box.
            page (int, optional): The page number. Defaults to 0.

        Raises:
            ValueError: If the coordinates are invalid.
            ValueError: If the page number is invalid.

        Returns:
            str: Extracted text that intersects with the bounding box.
        """
        self.__check_coordinates(x0, y0, x1, y1)
        self.__check_page_number(page)

        spans: list[Span] = self.extract_spans(x0, y0, x1, y1, page)
        extracted_text: list[str] = []

        for span in spans:
            extracted_text.append(span.to_text())
        return "".join(extracted_text)



import pytest

from tests.test_benchmark import perform_memory_test, perform_speed_test


@pytest.mark.skip()
def test_speed_bible(bible_file_name):
    perform_speed_test(bible_file_name, 100)


@pytest.mark.skip()
def test_memory_bible(bible_file_name):
    perform_memory_test(bible_file_name, 1300)



import math
from collections.abc import Generator
from hashlib import md5
from typing import Any

from pdfminer.layout import LTChar, LTPage, LTText, LTTextContainer, LTTextLine

from .data.classes import HotCharacter, PageResult
from .helpers.nanoid import generate_nano_id
from .span_map import SpanMap
from .sparse_matrix import SparseMatrix
from .trie import Trie


class MemoryMap:
    def __init__(self) -> None:
        """Initialize the MemoryMap. 2D Matrix representation of a PDF Page.

        Args:
            width (int): The width (max columns) of a page.
            height (int) The height (max rows) of a page.
        """
        self.text_trie = Trie()
        self.span_map = SpanMap()
        self.width: int = 0
        self.height: int = 0

    def build_memory_map(self) -> None:
        """Build the memory map based on width and height.

        The memory map is a SparseMatrix representation of the PDF.
        """
        self.memory_map = SparseMatrix()

    def __get_page_spans(self, page: LTPage) -> Generator[LTTextContainer[Any], None, None]:
        for element in page:
            if isinstance(element, LTTextContainer):
                yield element

    def load_memory_map(self, page: LTPage, drop_duplicate_spans: bool = True) -> None:
        """Load memory map data from an XML page.

        Args:
            page (str): The XML page data.
            drop_duplicate_spans (bool): Drop spans that are duplicates (example: on top of each other)

        Returns:
            None
        """
        char_hot_characters: list[tuple[str, HotCharacter]] = []
        spans: Generator[LTTextContainer[Any], None, None] = self.__get_page_spans(page)
        seen_span_hashes: list[str] = []
        for span in spans:
            span_hash = md5(str(seen_span_hashes).encode(), usedforsecurity=False).hexdigest()
            if span_hash in seen_span_hashes:
                continue
            seen_span_hashes.append(span_hash)
            span_id = generate_nano_id(size=10)
            for line in span:
                if not isinstance(line, LTTextLine):
                    continue
                for character in line:
                    if isinstance(character, (LTChar, LTText)) and (
                        hasattr(character, "x0")
                        and hasattr(character, "x1")
                        and hasattr(character, "y0")
                        and hasattr(character, "y1")
                    ):
                        char_c = character.get_text()
                        x0 = round(character.x0)
                        x1 = round(character.x1)
                        y0 = round(page.height - character.y0)
                        hot_character = HotCharacter(
                            value=char_c,
                            x=x0,
                            y=y0,
                            x_end=x1,
                            span_id=span_id,
                        )
                        self.memory_map.insert(value=char_c, row_idx=y0, column_idx=x0)
                        char_hot_characters.append((
                            char_c,
                            hot_character,
                        ))
        # Insert into Trie and Span Maps
        _hot_character: HotCharacter
        for char_c, _hot_character in char_hot_characters:
            self.text_trie.insert(char_c, _hot_character)
            if _hot_character.span_id:
                self.span_map[_hot_character.span_id] = _hot_character
        self.width = math.ceil(page.width)
        self.height = math.ceil(page.height)

    def extract_text_from_bbox(self, x0: int, x1: int, y0: int, y1: int) -> str:
        """Extract text within a specified bounding box.

        Args:
            x0 (int): Left x-coordinate of the bounding box.
            x1 (int): Right x-coordinate of the bounding box.
            y0 (int): Bottom y-coordinate of the bounding box.
            y1 (int): Top y-coordinate of the bounding box.

        Returns:
            str: Extracted text within the bounding box.
        """
        extracted_text: str = ""
        for row in range(max(y0, 0), min(y1, self.memory_map.rows - 1) + 1):
            row_text: str = ""
            row_text = "".join([
                self.memory_map.get(row_idx=row, column_idx=col)
                for col in range(max(x0, 0), min(x1, self.memory_map.columns - 1) + 1)
            ])
            if row_text:
                extracted_text += row_text + "\n"

        return extracted_text

    def find_text(self, query: str) -> tuple[list[str], PageResult]:
        """Find text within the memory map.

        Args:
            query (str): The text to search for.

        Returns:
            list: List of found text coordinates.
        """
        found_text = self.text_trie.search_all(query)
        return found_text



import os
from unittest.mock import patch

import pytest
from pdfminer.pdfdocument import PDFPasswordIncorrect
from pdfminer.pdfparser import PDFSyntaxError

from hotpdf import HotPdf
from hotpdf.data.classes import ElementDimension
from hotpdf.memory_map import MemoryMap
from hotpdf.utils import get_element_dimension


def test_load(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)


def test_load_constructor(valid_file_name):
    hotpdf_obj = HotPdf(valid_file_name)
    assert len(hotpdf_obj.pages) > 0


def test_load_bytes(valid_file_name):
    with open(valid_file_name, "rb") as f:
        hot_pdf_object = HotPdf()
        hot_pdf_object.load(f)


def test_load_locked(locked_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(locked_file_name, password="hotpdfiscool")
    page = hot_pdf_object.extract_page_text(page=0)
    assert len(page) > 500


def test_load_locked_bytes(locked_file_name):
    hot_pdf_object = HotPdf()
    with open(locked_file_name, "rb") as f:
        hot_pdf_object.load(f, password="hotpdfiscool")
    page = hot_pdf_object.extract_page_text(page=0)
    assert len(page) > 500


def test_load_locked_wrong_psw(locked_file_name):
    hot_pdf_object = HotPdf()
    with pytest.raises(PDFPasswordIncorrect):
        hot_pdf_object.load(locked_file_name, password="defenitelythewrongpassword")


def test_load_locked_wrong_psw_bytes(locked_file_name):
    hot_pdf_object = HotPdf()
    with open(locked_file_name, "rb") as f, pytest.raises(PDFPasswordIncorrect):
        hot_pdf_object.load(f, password="defenitelythewrongpassword")


def test_load_locked_no_psw(locked_file_name):
    hot_pdf_object = HotPdf()
    with pytest.raises(PDFPasswordIncorrect):
        hot_pdf_object.load(locked_file_name)


def test_load_locked_no_psw_bytes(locked_file_name):
    hot_pdf_object = HotPdf()
    with open(locked_file_name, "rb") as f, pytest.raises(PDFPasswordIncorrect):
        hot_pdf_object.load(f)


def test_load_invalid(invalid_file_name):
    with pytest.raises(PDFSyntaxError):
        HotPdf(invalid_file_name)


def test_full_text(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    text_first_page = hot_pdf_object.extract_page_text(page=0)
    # Not blank extraction
    assert len(text_first_page) > 500


def test_full_text_bytes(valid_file_name):
    with open(valid_file_name, "rb") as f:
        hot_pdf_object = HotPdf()
        hot_pdf_object.load(f)
        text_first_page = hot_pdf_object.extract_page_text(page=0)
        # Not blank extraction
        assert len(text_first_page) > 500


def test_pages_length(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    pages = hot_pdf_object.pages
    assert len(pages) == 1


def test_extraction(valid_file_name):
    WORD = "DEGREE"
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    occurences = hot_pdf_object.find_text(WORD)
    assert len(occurences) == 1
    for page_num, _ in occurences.items():
        assert page_num == 0

    element = get_element_dimension(occurences[0][0])
    extracted_text = hot_pdf_object.extract_text(
        x0=element.x0,
        y0=element.y0,
        x1=element.x1,
        y1=element.y1,
    )
    extracted_text = extracted_text.strip("\n").strip()
    assert WORD in extracted_text


def test_full_span_extraction(valid_file_name):
    WORD = "EXPERIENCE"

    to_check = {
        "EXPERIENCE ": False,
        "VOLUNTEER EXPERIENCE OR LEADERSHIP": False,
    }

    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    occurences = hot_pdf_object.find_text(WORD, take_span=True)

    # Only 1 page
    assert len(occurences) == 1
    # Should be page 0
    for page_num, _ in occurences.items():
        assert page_num == 0

    for occurence in occurences[0]:
        dimension = get_element_dimension(occurence)
        text_extracted = hot_pdf_object.extract_text(
            x0=dimension.x0,
            y0=dimension.y0,
            x1=dimension.x1,
            y1=dimension.y1,
        )
        if text_extracted in to_check:
            to_check[text_extracted] = True
    assert all([to_check.values()])


def test_full_span_extraction_sorted(valid_file_name):
    WORD = "EXPERIENCE"

    checks = [
        "EXPERIENCE",
        "VOLUNTEER EXPERIENCE OR LEADERSHIP",
    ]

    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    occurences = hot_pdf_object.find_text(WORD, take_span=True, sort=True)

    # Only 1 page
    assert len(occurences) == 1
    # Should be page 0
    for page_num, _ in occurences.items():
        assert page_num == 0
    for i in range(len(checks)):
        occurence = occurences[0][i]
        dimension = get_element_dimension(occurence)
        text_extracted = hot_pdf_object.extract_text(
            x0=dimension.x0,
            y0=dimension.y0,
            x1=dimension.x1,
            y1=dimension.y1,
        )
        text_extracted = text_extracted.strip("\n").strip()
        assert text_extracted == checks[i]


def test_non_existent_file_path(non_existent_file_name):
    with pytest.raises(FileNotFoundError):
        hot_pdf_object = HotPdf()
        hot_pdf_object.load(non_existent_file_name)


def test_blank_pdf(blank_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(blank_file_name)
    len_pages = len(hot_pdf_object.pages)
    assert all([len(hot_pdf_object.extract_page_text(page=i).strip("\n").strip()) == 0 for i in range(len_pages)])


def test_row_index_greater_than_rows_of_memory_map(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    pages = hot_pdf_object.pages
    assert pages[0].memory_map.get(row_idx=9999, column_idx=100) is not None
    assert pages[0].memory_map.get(row_idx=9999, column_idx=100) == ""


def test_col_index_greater_than_columns_of_memory_map(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    pages = hot_pdf_object.pages
    assert pages[0].memory_map.get(row_idx=100, column_idx=9999) is not None
    assert pages[0].memory_map.get(row_idx=100, column_idx=9999) == ""


@pytest.mark.parametrize("page", [-1, -2, 99])
def test_invalid_page_number(valid_file_name, page):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)

    with pytest.raises(ValueError, match="Invalid page number"):
        hot_pdf_object.find_text("Test", pages=[page])

    with pytest.raises(ValueError, match="Invalid page number"):
        hot_pdf_object.extract_text(x0=0, y0=1, x1=100, y1=5, page=page)


@pytest.mark.parametrize(
    "coordinates",
    [
        [-1, 0, 0, 0],
        [0, -1, 0, 0],
        [0, 0, -1, 0],
        [0, 0, 0, -1],
    ],
)
def test_extract_invalid_coordinates(valid_file_name, coordinates):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)

    with pytest.raises(ValueError, match="Invalid coordinates"):
        hot_pdf_object.extract_text(x0=coordinates[0], y0=coordinates[1], x1=coordinates[2], y1=coordinates[3])


def test_get_spans(valid_file_name):
    INCOMPLETE_WORD = "EXPERIENCE"
    NON_EXISTENT_WORD = "BLAH"
    # hotpdf should return the following span values

    to_check = {
        "EXPERIENCE ": False,
        "VOLUNTEER EXPERIENCE OR LEADERSHIP": False,
    }
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)

    occurences = hot_pdf_object.find_text(INCOMPLETE_WORD, take_span=True)
    element_dimensions: list[ElementDimension] = []
    for _, page_num in enumerate(occurences):
        occurences_by_page = occurences[page_num]
        for occurence_by_page in occurences_by_page:
            element_dimension = get_element_dimension(occurence_by_page)
            full_spans_in_bbox = hot_pdf_object.extract_spans(
                x0=element_dimension.x0,
                y0=element_dimension.y0,
                x1=element_dimension.x1,
                y1=element_dimension.y1,
            )
            assert len(full_spans_in_bbox) == 1
            element_dimensions.append(full_spans_in_bbox[0].get_element_dimension())

    for span_dim in element_dimensions:
        text_extracted = hot_pdf_object.extract_text(x0=span_dim.x0, y0=span_dim.y0, x1=span_dim.x1, y1=span_dim.y1)
        if text_extracted in to_check:
            to_check[text_extracted] = True

    assert all([to_check.values()])

    # Test Non Existent Word
    occurences = hot_pdf_object.find_text(NON_EXISTENT_WORD)
    assert occurences == {0: []}


@pytest.mark.parametrize("first_page, last_page", [(1, 1), (1, 2)])
def test_extract_page_range(multiple_pages_file_name, first_page, last_page):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(multiple_pages_file_name, first_page=first_page, last_page=last_page)
    pages = hot_pdf_object.pages
    assert len(pages) == last_page - first_page + 1


@pytest.mark.parametrize("first_page, last_page", [(-1, 1), (20, 10)])
def test_extract_page_range_exception(multiple_pages_file_name, first_page, last_page):
    hot_pdf_object = HotPdf()
    with pytest.raises(ValueError, match="Invalid page range"):
        hot_pdf_object.load(multiple_pages_file_name, first_page=first_page, last_page=last_page)


def test_no_spans_in_xml_file_extract_spans(valid_file_name):
    hot_pdf_object = HotPdf()
    with patch.object(MemoryMap, "_MemoryMap__get_page_spans") as get_page_spans:
        hot_pdf_object.load(valid_file_name)
        get_page_spans.return_value = None
        # Test No spans
        assert (
            hot_pdf_object.extract_spans(
                x0=0,
                y0=0,
                x1=100,
                y1=100,
            )
            == []
        )


def test_find_text_multiple_pages(multiple_pages_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(multiple_pages_file_name)
    occurences = hot_pdf_object.find_text(query="God", pages=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11])
    assert len(occurences) == 11


def test_find_spans_unique(valid_file_name):
    hot_pdf_object = HotPdf(valid_file_name)
    occurences = hot_pdf_object.find_text("EXPERIENCE")
    assert len(occurences[0]) == 2


def test_extract_spans(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    spans = hot_pdf_object.extract_spans(0, 0, 1000, 1000)
    to_check = {
        "PDF ": False,
        "THE BEST PDF PARSING LIBRARY TO EVER EXIST(DEBATABLE) ": False,
        "HOTPDF ": False,
    }
    all_text: str = ""
    for span in spans:
        span_text: str = span.to_text()
        if span_text in to_check:
            to_check[span_text] = True
        all_text += span_text
    assert all(to_check.values()), "Not all spans were extracted"
    assert len(all_text) == 608 if os.name == "nt" else 728


def test_span_has_no_characters(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    spans = hot_pdf_object.extract_spans(0, 0, 1000, 1000)
    span_1 = spans[0]
    assert span_1.get_element_dimension()
    span_1.characters = None
    with pytest.raises(ValueError, match="Span has no characters"):
        span_1.get_element_dimension()


def test_extract_spans_text(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    text = hot_pdf_object.extract_spans_text(0, 0, 1000, 1000)

    assert len(text) == 608 if os.name == "nt" else 728


def test_CONSISTENCY(valid_file_name):
    hot_pdf_object = HotPdf()
    hot_pdf_object.load(valid_file_name)
    page_text = hot_pdf_object.extract_page_text(0)
    bbox_text = hot_pdf_object.extract_text(0, 0, 1000, 1000)
    assert page_text == bbox_text


def test_load_pages_defined(multiple_pages_file_name):
    hotpdf_object = HotPdf()
    hotpdf_object.load(multiple_pages_file_name, first_page=1, last_page=1)
    assert len(hotpdf_object.pages) == 1

    hotpdf_object_1 = HotPdf(multiple_pages_file_name, first_page=1, last_page=3)
    assert len(hotpdf_object_1.pages) == 3
